@@ -59,7 +59,7 @@ struct ScannerPrivate
 {
     GAsyncQueue *scan_queue;
     gboolean running;
-    GThread *thread;
+    GThread *thread, *detect_thread;
 };
 
 G_DEFINE_TYPE (Scanner, scanner, G_TYPE_OBJECT);
@@ -137,7 +137,7 @@ poll_for_devices (Scanner *scanner)
     GList *devices = NULL;
 
     g_debug ("sane_get_devices ()");
-    status = sane_get_devices (&device_list, SANE_FALSE);
+    status = sane_get_devices (&device_list, SANE_TRUE);
     if (status != SANE_STATUS_GOOD) {
         g_warning ("Unable to get SANE devices: %s", sane_strstatus(status));
         return;
@@ -250,7 +250,6 @@ static void
 log_option (SANE_Int index, const SANE_Option_Descriptor *option)
 {
     GString *string;
-    SANE_String_Const *string_iter;
     SANE_Word i;
     SANE_Int cap;
     
@@ -388,10 +387,22 @@ log_option (SANE_Int index, const SANE_Option_Descriptor *option)
 }
 
 
+static gpointer
+detect_thread (Scanner *scanner)
+{
+    while (scanner->priv->running) {
+        poll_for_devices (scanner);
+        g_usleep (1000);
+    }
+    
+    return NULL;
+}
+
+
 static gpointer
 scan_thread (Scanner *scanner)
 {
-    ScanRequest *request;
+    ScanRequest *request = NULL;
     SANE_Status status;
     SANE_Handle handle = NULL;
     SANE_Parameters parameters;
@@ -402,6 +413,8 @@ scan_thread (Scanner *scanner)
     SANE_Byte *data = NULL;
     SANE_Int version_code;
     gboolean done = FALSE;
+    GError *error = NULL;
+    gchar *open_device = NULL;
 
     g_debug ("sane_init ()");
     status = sane_init (&version_code, NULL);
@@ -414,11 +427,17 @@ scan_thread (Scanner *scanner)
              SANE_VERSION_MINOR(version_code),
              SANE_VERSION_BUILD(version_code));
 
+    // NOTE: This assumes that sane_get_devices is thread safe
+    scanner->priv->detect_thread = g_thread_create ((GThreadFunc) detect_thread, scanner, TRUE, &error);
+    if (error) {
+        g_critical ("Unable to create thread: %s", error->message);
+        g_error_free (error);
+    }
+
     while (scanner->priv->running) {
         /* Look for requests */
         if (state == STATE_IDLE) {
-            GTimeVal timeout = { 1, 0 };
-            request = g_async_queue_timed_pop (scanner->priv->scan_queue, &timeout);
+            request = g_async_queue_pop (scanner->priv->scan_queue);
         } else if (state != STATE_CLOSE) {
             if (g_async_queue_length (scanner->priv->scan_queue) > 0)
                 state = STATE_CLOSE;
@@ -427,14 +446,26 @@ scan_thread (Scanner *scanner)
         /* Interrupted */
         if (!scanner->priv->running)
             break;
-        
+
         switch (state) {
         case STATE_IDLE:
-            if (request == NULL) {
-                poll_for_devices (scanner);
-            } else if (request->device) {
-                g_debug ("sane_open (\"%s\")", request->device);
-                status = sane_open (request->device, &handle);
+            if (request->device) {
+                /* Close existing device */
+                if (open_device && strcmp (open_device, request->device) != 0) {
+                    g_debug ("sane_close ()");
+                    sane_close (handle);
+                    handle = NULL;
+                    open_device = NULL;
+                }
+                
+                if (open_device) {
+                    status = SANE_STATUS_GOOD;
+                }
+                else {
+                    g_debug ("sane_open (\"%s\")", request->device);
+                    status = sane_open (request->device, &handle);
+                }
+
                 if (status != SANE_STATUS_GOOD) {
                     g_warning ("Unable to get open device: %s", sane_strstatus (status));
                     emit_signal (scanner, SCAN_FAILED,
@@ -444,6 +475,7 @@ scan_thread (Scanner *scanner)
                     state = STATE_CLOSE;
                 }
                 else {
+                    open_device = g_strdup (request->device);
                     state = STATE_GET_OPTION;
                     option_index = 0;
                     pass_number = 0;
@@ -451,6 +483,9 @@ scan_thread (Scanner *scanner)
                     notified_page = -1;
                 }
             }
+            else {
+                state = STATE_CLOSE;
+            }
             break;
 
         case STATE_GET_OPTION:
@@ -637,11 +672,6 @@ scan_thread (Scanner *scanner)
 
         case STATE_CLOSE:
             emit_signal (scanner, IMAGE_DONE, NULL);
-            if (handle) {
-                g_debug ("sane_close ()");
-                sane_close (handle);
-            }
-            handle = NULL;
             g_free (data);
             data = NULL;
             g_free (request->device);
@@ -680,7 +710,7 @@ scanner_start (Scanner *scanner)
 
 void
 scanner_scan (Scanner *scanner, const char *device, const char *source,
-              gint dpi, const char *scan_mode, gint depth, gboolean multi_page)
+              gint dpi, ScanMode scan_mode, gint depth, gboolean multi_page)
 {
     ScanRequest *request;
 
@@ -691,8 +721,19 @@ scanner_scan (Scanner *scanner, const char *device, const char *source,
     if (source)
         request->source = g_strdup (source);
     request->dpi = dpi;
-    if (scan_mode)
-        request->scan_mode = g_strdup (scan_mode);
+    switch (scan_mode) {
+    case SCAN_MODE_COLOR:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_COLOR);
+        break;
+    case SCAN_MODE_GRAY:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_GRAY);
+        break;
+    case SCAN_MODE_LINEART:
+        request->scan_mode = g_strdup (SANE_VALUE_SCAN_MODE_LINEART);
+        break;
+    default:
+        break;
+    }
     request->depth = depth;
     request->multi_page = multi_page;
     g_async_queue_push (scanner->priv->scan_queue, request);
@@ -702,7 +743,7 @@ scanner_scan (Scanner *scanner, const char *device, const char *source,
 void
 scanner_cancel (Scanner *scanner)
 {
-    scanner_scan (scanner, NULL, NULL, 0, NULL, 0, FALSE);
+    scanner_scan (scanner, NULL, NULL, 0, SCAN_MODE_DEFAULT, 0, FALSE);
 }
 
 
@@ -713,6 +754,8 @@ void scanner_free (Scanner *scanner)
     g_async_queue_push (scanner->priv->scan_queue, "");
     if (scanner->priv->thread)
         g_thread_join (scanner->priv->thread);
+    if (scanner->priv->detect_thread)
+        g_thread_join (scanner->priv->detect_thread);    
 
     g_async_queue_unref (scanner->priv->scan_queue);
     g_object_unref (scanner);
